from awsee.messages import Messages
import boto3
import datetime, time
import pytz
from dateutil import tz
from datetime import timedelta, tzinfo
from awsee.logmanager import LogManager
from awsee.style import Style
from awsee.utils import Utils
from botocore.exceptions import ClientError


class AwsServices:

    def __init__(self):
        pass

    def botoSession(self, profile):
        if not profile:
           return boto3.Session()
        return boto3.Session(profile_name=profile)

    def getAccountOwner(self, profile=None):
        sts = self.botoSession(profile).client("sts")
        return sts.get_caller_identity()

    def getMFADevices(self, profile=None):
        iam = self.botoSession(profile).client("iam")
        return iam.list_mfa_devices()

    def botoSessionWithToken(self, stsToken):
        session = boto3.session.Session(
            aws_access_key_id=stsToken['Credentials']['AccessKeyId'],
            aws_secret_access_key=stsToken['Credentials']['SecretAccessKey'],
            aws_session_token=stsToken['Credentials']['SessionToken']
        )
        return session
    
    def startSessionWithTokenAndAssumeRole(self, _stsToken, roleArn):
        stsToken = self.botoSessionWithToken(_stsToken).client('sts').assume_role(
            RoleArn=roleArn,
            RoleSessionName="mysession",
            DurationSeconds=3600
        )
        self._convertsUTCDateExpirationToLocaZoneDate(stsToken)
        return stsToken


    def assumeRole(self, roleArn, profile=None, mfaSerial=None, mfaToken=None):
        try:
            stsToken = None
            if mfaSerial and mfaToken:
                # In case a MFA Token is necessary to assume the role
                stsToken = self.botoSession(profile).client('sts').assume_role(
                    RoleArn=roleArn,
                    RoleSessionName="mysession",
                    DurationSeconds=3600,
                    SerialNumber=mfaSerial,
                    TokenCode=mfaToken
                )
            else:
                # In case the active AWS Session does not need a MFA Role 
                stsToken = self.botoSession(profile).client('sts').assume_role(
                    RoleArn=roleArn,
                    RoleSessionName="mysession",
                    DurationSeconds=3600
                )
        except ClientError as e:
            if e.response['Error']['Code'] == "AccessDenied":
                msg = e.response['Error']['Message']
                if not mfaSerial:
                    mfaSerial = ""
                Messages.showError("Access Denied!",f"{msg}\n    {mfaSerial}")
                return None
            else:
                print(e.response['Error']['Code'])
                raise e
        self._convertsUTCDateExpirationToLocaZoneDate(stsToken)
        return stsToken


    def _convertsUTCDateExpirationToLocaZoneDate(self, stsToken):
        localDate = Utils.parseUTCDateToLocalZoneDate(stsToken["Credentials"]["Expiration"])
        stsToken["Credentials"]["Expiration"] = localDate

    def getSessionToken(self, mfaSerial, mfaToken, profile=None, durationSeconds=3600):
        try:
            session = self.botoSession(profile)
            stsToken = session.client('sts').get_session_token(
                DurationSeconds=durationSeconds,
                SerialNumber=mfaSerial,
                TokenCode=mfaToken,
            )
        except ClientError as e:
            if e.response['Error']['Code'] == "AccessDenied":
                msg = e.response['Error']['Message']
                Messages.showError("Access Denied!",f"{msg}\n    {mfaSerial}")
                return None
            else:
                print(e.response['Error']['Code'])
                raise e
        self._convertsUTCDateExpirationToLocaZoneDate(stsToken)
        return stsToken
        
        # For Temporary Tests
        # stsToken = {
        #     "AccessKeyId": "ASIAV7NRRSDFSFFGGQSP24HZ",
        #     "SecretAccessKey": "QXsks0IQyTnS9qkUntDtrukfsldhyfnwlefGB8ztFZ7",
        #     "SessionToken": "FwoGZXIvYXdzEJ3//////////wEaDDVEVIvDLPKruvdC4CKFAQ7hnQi0eWG7pjprKnsmFNAnvgMtP4foPy6KygfDcBrEExJdWxnL0S4ok=",
        #     "Expiration": datetime.datetime.now()
        # }
        # return stsToken

        
        



# coding=utf-8
#
# Author: Ualter Otoni Pereira
# ualter.junior@gmail.com
#
import os
import logging
import sys
import re
import atexit 
from os.path import expanduser
import signal
import datetime
from datetime import timezone

from shutil import which
from tinydb import Query, where
from arnparse import arnparse
from awsee.messages import Messages
from awsee.style import Style
from awsee.emoticons import Emoticons
from awsee.utils import Utils
from awsee.tableargs import TableArgs
from awsee.prettytable import PrettyTable
from awsee.awsservices import AwsServices
from awsee.logmanager import LogManager
from awsee.preferences import Preferences
from awsee.functions import Functions
from awsee.rolemanager import RoleManager
from awsee.mfamanager import MFAManager
from awsee.repository import Repository, CredentialsRepository, ConfigRepository, RoleRepository, MfaRepository
from awsee.windowsenv import WindowsEnv
from awsee.sessionmanager import SessionManager
from awsee.profilemanager import ProfileManager

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.DEBUG)
FILE_LOG = expanduser("~") + "/.awsee/" + "log"

signal.signal(signal.SIGINT, signal.SIG_DFL)

class AwsSeePy:
    def __init__(self):
        myhomeFolder = expanduser("~") + "/.awsee/"
        if not os.path.exists(myhomeFolder):
           os.makedirs(myhomeFolder)

        self.mfaToken = None

        #LogManager().LOG.info("Starting...")
        self.credentialsRepository = CredentialsRepository()
        self.configRepository      = ConfigRepository()
        self.roleRepository        = RoleRepository()
        self.mfaRepository         = MfaRepository()
        self.preferences           = Preferences()
        self.executeFunction       = None
    
    def loadConfigProfiles(self):
        fileConfig  = expanduser("~") + "/.aws/config"

        if not os.path.exists(fileConfig):
            LogManager().LOG.warning("No file {} was found".format(fileConfig))
            return None
       
        configProfiles           = {}
        configProfiles["config"] = []
        currentProfileName = None 
        keys               = {}
        with open(fileConfig,'r') as fcfg:
            for line in fcfg.readlines():
                line = line.lstrip().rstrip()
                if line.lstrip().rstrip().startswith("["):
                   if currentProfileName:
                       configProfiles["config"].append({
                           "profile": currentProfileName,
                           "configurations": keys
                       })
                       keys = {}
                   # Read Current    
                   currentProfileName = Utils.cleanBreakLines(line).replace("[","").replace("]","")     
                else:
                   if Utils.cleanBreakLines(line).lstrip().rstrip() != "":
                      line = Utils.cleanBreakLines(re.sub('\s',"",line))
                      kv   = line.split("=")
                      keys[kv[0]] = kv[1]
            if currentProfileName:
               configProfiles["config"].append({
                   "profile": currentProfileName,
                   "configurations": keys
               })

        # Check if needs updates/inserts
        for c in configProfiles["config"]:
            dbRecordProfile = self.configRepository.findByProfile(c["profile"])
            # Insert the new one
            if not dbRecordProfile:
                self.configRepository.insert(c)
            else:
                # Update if needed
                update = False
                for config in c["configurations"]:
                    if config not in dbRecordProfile or \
                       c["configurations"][config] != dbRecordProfile["configurations"][config]:
                       update = True
                if update:
                   self.configRepository.update(c, c["profile"])

    def loadProfiles(self):
        fileCredentials = expanduser("~") + "/.aws/credentials"

        if not os.path.exists(fileCredentials):
            LogManager().LOG.warning("No file {} was found".format(fileCredentials))
            Messages.showWarning("Credentials Not Found!",f" The file {Style.IBLUE}/.aws/credentials{Style.GREEN} was not found, did you install and configure your {Style.IBLUE}awscli{Style.GREEN}?")
            sys.exit()

        currentProfileName                   = None
        accessKey                            = None
        secretKey                            = None
        credentialsProfiles                  = {}
        credentialsProfiles["credentials"]   = []
        with open(fileCredentials,'r') as fcred:
            for line in fcred.readlines():
                if line.lstrip().rstrip().startswith("["):
                   if currentProfileName:
                       defaultProfile = False
                       if currentProfileName == self.preferences.defaultProfile:
                           defaultProfile = True
                       # SAVE Previous
                       credentialsProfiles["credentials"].append({
                           "profile": currentProfileName,
                           "accessKey": accessKey,
                           "secretKey": secretKey,
                           "default": defaultProfile
                       })
                   # Read Current    
                   currentProfileName = Utils.cleanBreakLines(line).replace("[","").replace("]","") 
                else:
                   if Utils.cleanBreakLines(line).lstrip().rstrip() != "":
                      line = Utils.cleanBreakLines(re.sub('\s',"",line))
                      kv   = line.split("=")
                      if   "aws_access_key_id" in kv[0].lower():
                         accessKey = kv[1]
                      elif "aws_secret_access_key" in kv[0].lower():
                         secretKey = kv[1]
            if currentProfileName:
                defaultProfile = False
                if currentProfileName == self.preferences.defaultProfile:
                    defaultProfile = True
                # SAVE Previous
                credentialsProfiles["credentials"].append({
                    "profile": currentProfileName,
                    "accessKey": accessKey,
                    "secretKey": secretKey,
                    "default": defaultProfile
                })            

        # Check if needs updates/inserts
        awsServices = AwsServices()
        for c in credentialsProfiles["credentials"]:
            dbRecordProfile = self.credentialsRepository.findByProfile(c["profile"])
            # Insert the new one
            if not dbRecordProfile:
                account   = ""
                # Read Account and MFA Devices of the Profile at AWS
                try:
                    account = awsServices.getAccountOwner(c["profile"])["Account"]
                    for m in awsServices.getMFADevices(c["profile"])["MFADevices"]:
                        mfaRecord = self.mfaRepository.searchByQuery(Query().mfaDevice == m["SerialNumber"])
                        if not mfaRecord or len(mfaRecord) < 0:
                            self.mfaRepository.insert({
                                "profile": c["profile"],
                                "mfa-device": m["SerialNumber"],
                                "user-name": m["UserName"]
                            })
                except:
                    LogManager().LOG.warning(f"Not able to retrieve the MFA Devices for the {c['profile']}, probably some police might be blocking by the lack of MFA Token")
                c["account"] = account
                self.credentialsRepository.insert(c)
            else:
                # Check if needs to update the record (credentials change)
                if dbRecordProfile["accessKey"] != c["accessKey"]  or \
                   dbRecordProfile["secretKey"] != c["secretKey"]  or \
                   dbRecordProfile["default"]   != c["default"]:
                   # UPDATEIT
                   self.credentialsRepository.update(c, c["profile"])
                   
                # Try to Update its MFA Devices 
                # Removed to avoid remote connect all the time when the command --list is called
                #  TODO: Check to create a new function later, like --update-mfa-devices to explicitly update 
                #  the local credentials with MFA Devices remotely (if not blocked by polices)
                refreshMFA = False
                if refreshMFA:
                    try:
                        for m in awsServices.getMFADevices(c["profile"])["MFADevices"]:
                            mfaRecord = self.mfaRepository.searchByQuery(where('mfa-device') == m["SerialNumber"])
                            # Add a new MFA Device found at his/her Account (not in the MFA Device Repository yet)
                            if not mfaRecord or len(mfaRecord) < 0:
                                self.mfaRepository.insert({
                                    "profile": c["profile"],
                                    "mfa-device": m["SerialNumber"],
                                    "user-name": m["UserName"]
                                })
                    except:
                        pass

    def synchronize(self):
        print("")
        Messages.showStartExecution("Wait, synchronizing...")
        self.credentialsRepository.purge()
        self.configRepository.purge()
        self.loadProfiles()
        self.loadConfigProfiles()
        Messages.showStartExecution("Synchronization done!             ")
        print("")
    
    def listProfiles(self, more):
        profileManager = ProfileManager()
        profileManager.listProfiles(more)

    def addRole(self):
        roleManager = RoleManager()
        roleManager.addRole()

    def listRoles(self, filterByProfile):
        roleManager = RoleManager()
        roleManager.listRoles(filterByProfile)

    def removeRole(self, numberInList, roleName):
        roleManager = RoleManager()
        roleManager.removeRole(numberInList, roleName)
    
    def addMFADevice(self):
        mfaManager = MFAManager()
        mfaManager.addMfa()

    def listMFADevices(self, filterByProfile):
        mfaManager = MFAManager()
        mfaManager.listmfaDevices(filterByProfile)

    def removeRole(self, numberInList):
        mfaManager = MFAManager()
        mfaManager.removeRole(numberInList)

    def assumeRole(self, assumeRoleName, nameProfile=None, mfaToken=None, ):
        roleRecords = self.roleRepository.searchByQuery(where('role-name') == assumeRoleName)
        if not roleRecords or len(roleRecords) < 1:
           Messages.showWarning("Role not found!",f"     Role with the name {Style.IBLUE}{assumeRoleName}{Style.GREEN} was not found")
           sys.exit()

        role           = roleRecords[0]
        sessionManager = SessionManager()
        output, scriptFileWin, scriptFileBash = sessionManager.assumeRole(role, nameProfile, mfaToken)
        if output:
           output = Utils.removeCharsColors(output) if self.preferences.noColor else output
           print(output)

    def startAWSSessionWithMFAToken(self, nameProfile=None, mfaToken=None, assumeRole=None):
        if not nameProfile:
           # If not informed, use the Default Profile
           profiles = self.credentialsRepository.searchByQuery(where('default') == True)
           if not profiles or len(profiles) < 1:
               Messages.showWarning("Default profile not defined!")
               sys.exit()
           profile = profiles[0]
        else:
            # Get the Profile using the name informed
            profile = self.credentialsRepository.findByProfile(nameProfile)
            if not profile:
               Messages.showWarning(f"Profile {Style.IBLUE}{nameProfile}{Style.GREEN} not defined!")
               sys.exit()

        mfa = self.mfaRepository.findByProfile(profile['profile'])
        if not mfa:
           Messages.showWarning(f"MFA for the profile {profile['profile']} not found!")
           sys.exit()

        role = None
        if assumeRole:
            roles = self.roleRepository.searchByQuery(where('role-name') == assumeRole)
            if not roles or len(roles) < 1:
               Messages.showWarning(f"Role {assumeRole} not found!")
               sys.exit()
            role = roles[0]

        sessionManager = SessionManager()
        output, scriptFileWin, scriptFileBash = sessionManager.startAWSSessionWithMFA(profile, mfa, mfaToken, role)
        if output:
           output = Utils.removeCharsColors(output) if self.preferences.noColor else output
           print(output)
    
    def startAWSSession(self, nameProfile):
        profile = self.credentialsRepository.findByProfile(nameProfile)
        if not profile:
           Messages.showWarning(f"Profile {Style.IBLUE}{nameProfile}{Style.GREEN} not defined!")
           sys.exit()
        
        sessionManager = SessionManager()
        output, scriptFileWin, scriptFileBash = sessionManager.startAWSSession(profile)
        if output:
           output = Utils.removeCharsColors(output) if self.preferences.noColor else output
           print(output)
        

    def syntax(self, short=False):
        print("")
        functions = Functions()
        functions.showUsage(short)
        functions.showFunctions(short)
    
    def showInfo(self):
        sessionFound = False
        try:
            if os.environ['AWS_SEE']:
                if os.environ['AWS_SEE'] == "CLOSED":
                    Messages.showWarning("The AWSee Session was closed already!")
                    sys.exit()
                sessionFound = True
                account    = os.environ['AWS_ACCOUNT']
                user       = os.environ['AWS_USER']
                accessKey  = os.environ['AWS_ACCESS_KEY_ID']
                secretKey  = os.environ['AWS_SECRET_ACCESS_KEY']
                expiration = os.environ['AWS_EXPIRATION']
                role       = os.environ['AWS_ROLE']

                expDate = datetime.datetime.strptime(expiration,"%Y-%m-%d %H:%M:%S%z")
                diff    = (expDate - datetime.datetime.now(timezone.utc)).seconds / 60

                output = f"""\n{Style.IMAGENTA}#awsee{Style.GREEN}
=============================
     {Style.IMAGENTA}My AWSee Session{Style.GREEN}
=============================
{Style.GREEN}Account...........: {Style.IBLUE}{account}
{Style.GREEN}User..............: {Style.IBLUE}{user}
{Style.GREEN}Access Key........: {Style.IBLUE}{accessKey}
{Style.GREEN}Secret Key........: {Style.IBLUE}{secretKey}
{Style.GREEN}Expiration........: {Style.IBLUE}{expiration} ({Style.YELLOW}{diff:02.0f} {Style.IBLUE}minutes remaining)
{Style.GREEN}Role..............: {Style.IBLUE}{role}
                """
                output = Utils.removeCharsColors(output) if self.preferences.noColor else output
                print( output )
        except KeyError as e:
            LogManager().LOG.warning("KeyError while trying to read Environment variables")
            LogManager().LOG.warning("KeyError: " + str(e))
            sessionFound = False
        if not sessionFound:
            Messages.showWarning("No AWSee Session was started in this terminal!")
    
    def start(self):
        #executeFunction = False
        args = sys.argv
        if len(args) < 2:
           self.syntax()
           sys.exit()

        functions = Functions()
        if "-h" in args or "--help" in args:
            short = True if "-h" in args else False
            if len(args) > 2:
               f = functions.getFunctionByArgumentIdentifier(args[2])
               output = functions.showFunction(f)
               output = Utils.removeCharsColors(output) if self.preferences.noColor else output
               print(output)
            else:
               self.syntax(short)
            sys.exit()
        
        self.loadProfiles()
        self.loadConfigProfiles()

        args                 = args[1:]                          # Discard the first argument (our python function name)
        self.executeFunction = self.parseArgsFindFunction(args)  # Identify which function by the arguments/identifiers (-l --list, etc.)

        if  functions.INFO == self.executeFunction:
            self.showInfo()
        
        elif  functions.LIST_PROFILES == self.executeFunction:
            more = True if "full" in args or "-lf" in args else False
            self.listProfiles(more)

        elif functions.SYNC == self.executeFunction:
            self.synchronize()

        elif functions.ADD_ROLE == self.executeFunction:
            self.addRole()

        elif functions.LIST_ROLES == self.executeFunction:
            filterByProfile = None
            if len(args) > 1:
               filterByProfile = args[1]
            self.listRoles(filterByProfile)

        elif functions.REMOVE_ROLE == self.executeFunction:
            numberInList = -1
            roleName     = None
            if len(args) > 1:
               if Utils.isNumber(args[1]):
                   numberInList = int(args[1])
               else:
                   roleName = args[1]
            else:
                Messages.showWarning("Missing parameter! At least one of them must be informed:")
                output = Functions().showFunction(Functions.REMOVE_ROLE)
                output = Utils.removeCharsColors(output) if self.preferences.noColor else output
                print(output)
                sys.exit()
            self.removeRole(numberInList, roleName)

        elif functions.ADD_MFA == self.executeFunction:
             self.addMFADevice()

        elif functions.LIST_MFAS == self.executeFunction:
             filterByProfile = None
             if len(args) > 1:
                filterByProfile = args[1]
             self.listMFADevices(filterByProfile)

        elif functions.REMOVE_MFA == self.executeFunction:
            numberInList = -1
            if len(args) > 1:
               if Utils.isNumber(args[1]):
                   numberInList = int(args[1])
            else:
                Messages.showWarning("Missing parameter for Remove MFA...")
                output = Functions().showFunction(Functions.REMOVE_MFA)
                output = Utils.removeCharsColors(output) if self.preferences.noColor else output
                print(output)
                sys.exit()
            self.removeMFADevice(numberInList)

        elif functions.ASSUME_ROLE == self.executeFunction:
            if len(args) > 1:
               role = args[1]
               self.assumeRole(role,None,None)
            else:
               Messages.showWarning("Missing [ROLE] parameter...")
               output = Functions().showFunction(Functions.ASSUME_ROLE)
               output = Utils.removeCharsColors(output) if self.preferences.noColor else output
               print(output)
               sys.exit()

        elif functions.START_SESSION == self.executeFunction:
            if len(args) > 1:
                profile = args[1]
                self.startAWSSession(nameProfile=profile)
            else:
                Messages.showWarning("Missing [PROFILE] parameter...")
                output = Functions().showFunction(Functions.START_SESSION)
                output = Utils.removeCharsColors(output) if self.preferences.noColor else output
                print(output)
                sys.exit()

        elif functions.START_SESSION_MFA_TOKEN == self.executeFunction:
            # Parameters Already validated (if exists)
            mfaToken = args[0]
            profile  = args[2] if len(args) > 2 else None
            role     = args[4] if len(args) > 4 else None
            self.startAWSSessionWithMFAToken(nameProfile=profile, mfaToken=mfaToken, assumeRole=role)
        else:
             self.exitWithMessageInvalidUsage()

    
    def parseArgsFindFunction(self, args):
        functions = Functions()
        #
        # Looking for an auxiliar functions (Add Role, Add MFA, List Profiles, Remove Role, Sync, etc.)
        #
        for arg in args:
            func = functions.getFunctionByArgumentIdentifier(arg)
            if func:
               # Check if there's a MFA Token Informed, if it is, the funcion asked was not START_SESSION only, the argument -p must be the first one
               if func == functions.START_SESSION:
                  if args[0] == "-p":
                     return func
                  else:
                     continue
               # Check if there's a MFA Token Informed, if it is, the funcion asked was not ASSUME_ROLE, it is START_SESSION_MFA_TOKEN
               if func == functions.ASSUME_ROLE:
                  if args[0] == "-r":
                     return func
                  else:
                     continue 
               return func

        #
        # Check is command starting with a MFA Token (start AWS Session)
        #
        mfaToken = None
        if self.checkMFATokenValid(args[0]):
           mfaToken = args[0]
        else:
           Messages.showError(f"Invalid MFA Token {Style.GREEN}{args[0]}{Style.IMAGENTA}")
           sys.exit()
        # Check if a profile was passed: "-p [profile] or -p [profile] -r [role]"
        role = None
        if len(args) > 1:
           if len(args) == 4:
               self.exitWithMessageInvalidUsage()
           # Here, It should be -p [PROFILE]    
           elif len(args) == 3:
              if args[1] == "-p":
                 profile = args[2]
              elif args[1] != "-p":
                 self.exitWithMessageInvalidUsage()
           # Here, It should be -p [PROFILE] -r [ROLE]
           elif len(args) == 5:     
               if   args[1] == "-p":
                    profile = args[2]
               if   args[3] == "-p":
                    profile = args[4] 
               if args[1] == "-r":
                    role = args[2]  
               if args[3] == "-r":
                    role    = args[4]
               if (not profile or profile == "") or (not role or role == ""):
                   self.exitWithMessageInvalidUsage()
           else:
              Messages.showError(f"Missing parameter profile [-p profile], check syntax with {Style.GREEN}awsee -h{Style.RESET}")
              sys.exit()
        return functions.START_SESSION_MFA_TOKEN


    def exitWithMessageInvalidUsage(self):
        Messages.showError(f"Invalid usage!\n     Please, check syntax with {Style.GREEN}awsee -h{Style.RESET}")
        sys.exit() 

    def checkMFATokenValid(self, token):
        f = re.findall("[0-9]{6}$",token)
        if len(f) > 0:
            return True
        return False

@atexit.register 
def goodbye(): 
    pass
    #print("Bye!")
    

def main():
    """Start AwsSe"""
    try:
        awsee = AwsSeePy()
        awsee.start()
    except UnicodeEncodeError as e:
        if Utils.isWindows():
           output = f"\n{Style.IRED} --> {Style.IMAGENTA}UnicodeEncode Error{Style.GREEN}\n     If you are using GitBash at Windows\n     disable the use of emoticons at ~/.awsse/awsee.ini, {Style.IBLUE}emoticons-enabled = false{Style.RESET}\n\n"
           output = Utils.removeCharsColors(output) if Preferences().noColor else output
           print(output)

if __name__ == '__main__':
    main()







import os
from os.path import expanduser
import datetime
from awsee.logmanager import LogManager
from awsee.preferences import Preferences
from awsee.functions import Functions
from arnparse import arnparse
from awsee.awsservices import AwsServices
from awsee.messages import Messages
from awsee.style import Style
from awsee.emoticons import Emoticons
from awsee.utils import Utils
from tinydb import Query, where
from shutil import which
from awsee.repository import Repository, CredentialsRepository, ConfigRepository, RoleRepository, MfaRepository

class SessionManager:

    SCRIPTS_PREFIX_NAME_FILE = "set-awssession-"

    def __init__(self):
        self.credentialsRepository = CredentialsRepository()
        self.roleRepository        = RoleRepository()
        self.mfaRepository         = MfaRepository()
        self.preferences           = Preferences()

    # Start a Session and immediatly Assume a Role 
    def startCredentialsOwnerSessionWithMFA(self, mfaToken, profile, mfa):
        # TODO:
        pass 

    # Assume a Role
    def assumeRole(self, role, profile=None, mfaToken=None):
        aws = AwsServices()

        stsToken = aws.assumeRole(role['role-arn'], profile, mfaToken)
        if not stsToken:
            return None, None, None
        
        return self.createAWSSessionOnTerminalUser(stsToken["Credentials"], role['profile'], role['role-name'])
    
    # Open Session using MFA (default or other profile)
    def startAWSSession(self, profile):
        aws = AwsServices()

        stsToken = {}
        stsToken["AccessKeyId"]     = profile['accessKey']
        stsToken["SecretAccessKey"] = profile['secretKey']
        stsToken["SessionToken"]    = None
        stsToken["Expiration"]      = datetime.datetime.now() + datetime.timedelta(minutes=525600)

        return self.createAWSSessionOnTerminalUser(stsToken, profile['profile'], None)

    # Open Session (other profile - no need MFA)
    def startAWSSessionWithMFA(self, profile, mfa, mfaToken, role=None):
        aws = AwsServices()

        stsToken = aws.getSessionToken(mfaSerial=mfa['mfa-device'],mfaToken=mfaToken, profile=profile['profile'])
        if not stsToken:
            return None, None, None

        roleName = None
        if role:
           # If a Role was also passed, 
           # Initiate a session the the Profile given, and using it, immediately, create a Session with the Role passed
           roleName =  role['role-name']
           stsToken = aws.startSessionWithTokenAndAssumeRole(stsToken,role['role-arn'])

        return self.createAWSSessionOnTerminalUser(stsToken["Credentials"], profile['profile'], roleName)

    def createAWSSessionOnTerminalUser(self, stsToken, profile, role):
        accessKey      = stsToken["AccessKeyId"]
        secretKey      = stsToken["SecretAccessKey"]
        sessionToken   = stsToken["SessionToken"]
        expiration     = stsToken["Expiration"]
        output         = None
        scriptFileWin  = None
        scriptFileBash = None

        role            = "" if not role else "_" + role.replace(" ","_").lower()
        nameProfile     = profile.replace(" ","_")
        roleTitle       = "" if role == "" else f"Role: {role.replace('_','')}" 
        windowTitle     = f"AWSEE Profile: {nameProfile}  {roleTitle}"

        if Utils.isWindows():
            
            scriptFileWin   = f"{self.SCRIPTS_PREFIX_NAME_FILE}{nameProfile}{role}.bat"
            scriptFileBash  = f"{self.SCRIPTS_PREFIX_NAME_FILE}{nameProfile}{role}.sh"

            terminalCommand = self.preferences.windows.terminalCommand
            # Check if there's a Terminal Commmand Set
            if not terminalCommand or terminalCommand.lstrip().rstrip() == "":
                # Check CMDER is installed (then use it)
                if which("cmder") and os.environ['CMDER_ROOT'] != "":
                    outputWin, environmentsWin = self.createEnvironment(False, scriptFileWin, accessKey, secretKey, sessionToken, expiration, role.replace('_',''))
                    output = outputWin
                    os.system(f"start %CMDER_ROOT%\\vendor\\conemu-maximus5\\ConEmu.exe /icon \"%CMDER_ROOT%\\cmder.exe\" /title \"{windowTitle}\" /loadcfgfile \"%CMDER_ROOT%\\config\\ConEmu.xml\" /cmd cmd /k " + environmentsWin)
                    #os.system(f"start %CMDER_ROOT%\\vendor\\conemu-maximus5\\ConEmu.exe /icon \"%CMDER_ROOT%\\cmder.exe\" /title \"{windowTitle}\" /cmd cmd /k " + environmentsWin)
                else:
                    # Last option (windows default cmd)
                    if which("cmd"):
                        outputWin, environmentsWin = self.createEnvironment(False, scriptFileWin, accessKey, secretKey, sessionToken, expiration, role.replace('_',''))
                        output = outputWin
                        os.system("start cmd.exe /c cmd /k" + environmentsWin)
                    else:
                        msg = f"""\n{Style.GREEN}    Please set a {Style.IBLUE}terminal-command{Style.GREEN} at your {Style.IBLUE}~/.awsee/awsee.ini{Style.GREEN}, WINDOWS section in order \n    to open a Terminal ready to use with your AWS Session Token configured.
                        \n    Example: \n    {Style.IBLUE}[WINDOWS]\n    terminal-command = {Style.IGREEN}d:\Program Files\Git\git-bash.exe{Style.RESET}
                        """
                        Messages.showError("No Terminal Found (CMDER or CMD)",msg)
                        return None, None
            else:
                if "git-bash.exe" in terminalCommand or "sh.exe" in terminalCommand:
                    outputBash, environmentsBash = self.createEnvironment(True, scriptFileBash, accessKey, secretKey, sessionToken, expiration, role.replace('_',''))
                    output = outputBash

                    # Regarding the place where Python is running (not the Target Console to Open, configured at awsee.ini)...  
                    # Check if the Call to open a Shell Script Terminal were made inside a GitBash, if not... change to Windows Style (the Output Python window)
                    if not Utils.isRunningOnGitBash():
                        outputWin, _ = self.createEnvironment(False, scriptFileWin, accessKey, secretKey, sessionToken, expiration, role.replace('_',''))
                        output = outputWin

                if "git-bash.exe" in terminalCommand:
                    os.system(f"start {terminalCommand} --needs-console --no-hide --command=usr\\bin\\bash.exe --login -i -c \"sh -c 'source ./{environmentsBash}; exec sh'\"")
                elif "sh.exe" in terminalCommand:
                    os.system(f"start {terminalCommand} --login -i -l -c \"sh -c 'source ./{environmentsBash}; exec sh'\"")
                else:
                    msg = f"""\n{Style.GREEN}    Please check the {Style.IBLUE}terminal-command{Style.GREEN} set at your {Style.IBLUE}~/.awsee/awsee.ini{Style.GREEN}, WINDOWS section in order \n    to open a Terminal ready to use with your AWS Session Token configured.
                    \n    Example: \n    {Style.IBLUE}[WINDOWS]\n    terminal-command = {Style.IGREEN}d:\Program Files\Git\sh.exe{Style.RESET}
                    """
                    Messages.showError(f"Terminal Command set \"{terminalCommand}\" not supported (Use git-bash.exe, sh.exe or let it blank, for system default",msg)
                    return None, None
        else:
            # LINUX
            # TODO:
              print("export AWS_ACCESS_KEY")

        return output, scriptFileWin, scriptFileBash

    # Script to Help clean environment variables, etc. 
    def createCleanEnvironmentScript(self, scriptFile):
        if ".bat" in scriptFile:
            cleanFileName = "clean.bat"
            content = f"""@ECHO OFF
set AWS_ACCESS_KEY_ID=
set AWS_SECRET_ACCESS_KEY=
set AWS_SESSION_TOKEN=
set AWS_EXPIRATION=
set AWS_SEE=CLOSED
DEL {self.SCRIPTS_PREFIX_NAME_FILE}*
            """
        else:
            cleanFileName = "clean.sh"
            content = f"""unset AWS_ACCESS_KEY_ID
unset AWS_SECRET_ACCESS_KEY
unset AWS_SESSION_TOKEN
unset AWS_EXPIRATION
set AWS_SEE=CLOSED
rm {self.SCRIPTS_PREFIX_NAME_FILE}*
            """
        with open(f"{cleanFileName}",'w') as fout:
            fout.write(content)

    # Create Contents, Script File for Environments Variables and Clipboard (Win and Bash)
    def createEnvironment(self, isBashEnv, scriptFile, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        output           = self.createOutputStartSession(isBashEnv, scriptFile, accessKey, secretKey, sessionToken, expiration, roleTitle)
        initFileContent  = self.createContentScriptFileStartSession(isBashEnv, accessKey, secretKey, sessionToken, expiration, roleTitle)
        with open(f"{scriptFile}",'w') as fout:
            fout.write(initFileContent)
        environments = f"\"{scriptFile}\""
        self.addEnvironmentVariablesConfigurationToClipboard(isBashEnv,accessKey, secretKey, sessionToken, expiration, roleTitle)
        self.createCleanEnvironmentScript(scriptFile)
        return output, environments

    def createOutputStartSession(self, forBash, scriptFile, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        commandEnvSet = "set"
        executeScript = f"{scriptFile}"
        if forBash:
            commandEnvSet = "export"
            executeScript = f"source ./{scriptFile}"

        awsToken = f"{Style.GREEN}{commandEnvSet} AWS_SESSION_TOKEN{Style.WHITE}={Style.IBLUE}{sessionToken}\r" if sessionToken else ""
        awsRole  = f"{Style.GREEN}{commandEnvSet} AWS_ROLE{Style.WHITE}={Style.IBLUE}{roleTitle}\r" if roleTitle else ""

        output  = "\n"
        output += f" {Style.GREEN}======================="
        output += f"\n {Style.IMAGENTA}  >>> AWS SESSION <<<\n"
        output += f" {Style.GREEN}======================="
        output += f"""\n {Style.GREEN}set AWS_ACCESS_KEY_ID{Style.WHITE}={Style.IBLUE}{accessKey}
 {Style.GREEN}{commandEnvSet} AWS_SECRET_ACCESS_KEY{Style.WHITE}={Style.IBLUE}{secretKey}
 {awsToken}
 {Style.GREEN}{commandEnvSet} AWS_EXPIRATION{Style.WHITE}={Style.IBLUE}{expiration}
 {awsRole}

 {Style.IMAGENTA}  >>> Opening in another window...

 {Style.GREEN}or...
 {Style.GREEN}To setup this AWS Session at this same Window
 Hit [Shift + Insert] or execute the script:

 {Style.IBLUE}  >>> {executeScript}{Style.RESET}
        """
        return output

    def addEnvironmentVariablesConfigurationToClipboard(self, forBash, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        command = "set" if not forBash else "export"
        awsToken = f"{Style.GREEN}{command} AWS_SESSION_TOKEN{Style.WHITE}={Style.IBLUE}{sessionToken}\r" if sessionToken else ""
        # Add to Clipboard (SETs environmentsWin)
        forClipboard  = f"{command} AWS_ACCESS_KEY_ID={accessKey}\r\n"
        forClipboard += f"{command} AWS_SECRET_ACCESS_KEY={secretKey}\r\n"
        forClipboard += f"{awsToken}\r\n"
        forClipboard += f"{command} AWS_SEE=gG99USedshwRIVSe\r\n"
        if forBash:
           forClipboard += f"{command} AWS_EXPIRATION=\"{expiration}\"\r\n"
        else:
           forClipboard += f"{command} AWS_EXPIRATION={expiration}\r\n" 
        Utils.addToClipboard(forClipboard)

    def createContentScriptFileStartSession(self, forBash, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        initFileContent = self.createBashScriptFile(accessKey, secretKey, sessionToken, expiration, roleTitle) if forBash else \
                          self.createCMDscriptFile(accessKey, secretKey, sessionToken, expiration, roleTitle)
        return initFileContent

    def createCMDscriptFile(self, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        awsToken = f"set AWS_SESSION_TOKEN={sessionToken}" if sessionToken else ""
        roleEcho = f"ECHO.Role.........: {roleTitle}" if roleTitle else ""
        roleSet  = f"{roleTitle}" if roleTitle else " " 

        return f"""
        @ECHO OFF
        ECHO.#awsee
        ECHO.=============================
        ECHO.  AWSCLI Session Configured
        ECHO.=============================
        set AWS_ACCESS_KEY_ID={accessKey}
        set AWS_SECRET_ACCESS_KEY={secretKey}
        {awsToken}
        set AWS_EXPIRATION={expiration}

        for /F "delims=" %%i in ('aws sts get-caller-identity') do (
            SET "string=%%i"
            echo %%i | findstr /C:"Account">nul && (
                SET ACCOUNT=%%i
            )
            echo %%i | findstr /C:"Arn">nul && (
                SET USERARN=%%i
            )
        )
        set ACCOUNT=%ACCOUNT:Account=%
        set ACCOUNT=%ACCOUNT::=%
        set ACCOUNT=%ACCOUNT:"=%
        set ACCOUNT=%ACCOUNT:,=%
        set ACCOUNT=%ACCOUNT: =%

        set USERARN=%USERARN: =%
        set USERARN=%USERARN:~7%
        set USERARN=%USERARN:"=%
        REM set USERARN=%USERARN:Arn=%
        REM set USERARN=%USERARN:,=%
        

        set AWS_ACCOUNT=%ACCOUNT%
        set AWS_USER=%USERARN%
        set AWS_ROLE={roleSet}
        set AWS_SEE=gG99USedshwRIVSe

        REM Shortcuts
        DOSKEY ls=dir /W
        DOSKEY w=ECHO. ^&^& ECHO.#awsee ^&^& ECHO.============================= ^&^& ECHO.       MY AWS SESSION ^&^& ECHO.============================= ^&^& ECHO.Account...........:%ACCOUNT% ^&^& ECHO.User..............:%USERARN% ^&^& ECHO.Access Key........:%AWS_ACCESS_KEY_ID% ^&^& ECHO.Secret Key........:%AWS_SECRET_ACCESS_KEY% ^&^& ECHO.Expiration........:%AWS_EXPIRATION% ^&^& ECHO.Role..............:%AWS_ROLE%

        ECHO.Expires at...: {expiration:%Y-%m-%d %H:%M}
        ECHO.Account......: %ACCOUNT%
        ECHO.User.........: %USERARN%
        {roleEcho} 
        ECHO.
        """

    def createBashScriptFile(self, accessKey, secretKey, sessionToken, expiration, roleTitle=None):
        awsToken = f"export AWS_SESSION_TOKEN={sessionToken}" if sessionToken else ""
        roleEcho = f"echo Role.........: {roleTitle}" if roleTitle else " "
        roleSet  = f"{roleTitle}" if roleTitle else " " 
        
        return f"""
        #!/bin/sh
        echo 
        echo =============================
        echo   AWSCLI Session Configured
        echo =============================
        export AWS_ACCESS_KEY_ID={accessKey}
        export AWS_SECRET_ACCESS_KEY={secretKey}
        {awsToken}
        export AWS_EXPIRATION="{expiration}"

        vlr="$(aws sts get-caller-identity)"
        account=$(echo "$vlr" | grep '"Account"' | sed s/\\"//g | sed s/,//g | sed s/"Account:"//g)
        arn=$(echo "$vlr" | grep '"Arn"' | sed s/\\"//g | sed s/,//g | sed s/"Arn:"//g )

        export AWS_ACCOUNT=$account
        export AWS_USER=$arn
        export AWS_ROLE={roleSet}
        export AWS_SEE=gG99USedshwRIVSe

        echo Expires at...: "{expiration}"
        echo Account......: $account
        echo User.........: $arn
        {roleEcho}
        echo
        """
        